#!/usr/bin/env python
# Copyright 2016 Cumulus Networks, Inc
#
# ecmpcalc utility for the mlx platform

import subprocess
import sys
import os
import re
import ipaddr
from cumulus.ecmpcalc_util import asic2linux, asictrunkmbr2linux
try:
    from cumulus.mlx import mlx_open_connection
    from cumulus.mlx import mlx_close_connection
    from cumulus.mlx import mlx_get_uc_route
    from cumulus.mlx import mlx_get_ecmp
    from cumulus.mlx import mlx_get_neighbor
    from cumulus.mlx import mlx_get_interface
    from cumulus.mlx import mlx_get_fdb
    from cumulus.mlx import mlx_get_lag_member_ids
except ImportError:
    raise NotImplementedError


class RouteInfo:
    def __init__(self, vrf, dst, verbose=False):
        self.verbose = verbose
        self.vrf = vrf
        self.dst = dst
        self.isRoutable = False
        self.isEcmp = False
        self.rifs = []
        self.nhops = []
        self._routeLkup()

    def _routeLkup(self):
        dstAddr = ipaddr.IPAddress(self.dst.addrstr)

        # LPM on dstAddr
        rtTable = mlx_get_uc_route(version=dstAddr.version, vrid=None)
        #if self.verbose:
        #    sys.stderr.write('rtTable:\n  %s\n' % rtTable)
        net = 0
        ecmp_id = -1
        """
        Rt format:
        ('0 18.0.0.0/16', {'action': 'FORWARD', 'vrid': 0, 'prio': 0, 'next_hop_cnt': 3, 'erif': 6L, 'ecmp_id': 6, 'next_hops': ['11.0.0.25', '11.0.0.29', '11.0.0.37'], 'type': 'NEXT_HOP'})
        """
        for rt in rtTable.items():
            prefix = rt[0].split(' ')[1]
            newNet = ipaddr.IPNetwork(prefix)
            newEcmpId = rt[1].get('ecmp_id')
            if dstAddr in newNet and (ecmp_id == -1 or newNet > net):
                net = newNet
                ecmp_id = newEcmpId
                self.nhops = rt[1].get('next_hops', [])
                if self.verbose:
                    sys.stderr.write('matching rt: %s: %s\n' % (rt[0], rt[1]))
        if self.verbose:
            sys.stderr.write('located net %s, ecmp_id %d\n' % (net, ecmp_id))
        if ecmp_id != -1:
            self.isRoutable = True
        
        # Locate ecmp group
        if ecmp_id != -1 and len(self.nhops) > 1:
            ecmpTable = mlx_get_ecmp(ecmp_id)
            #if self.verbose:
            #    sys.stderr.write('ecmpTable:\n  %s\n' % ecmpTable)
            if ecmpTable:
                self.rifs = ecmpTable.get(ecmp_id, [])
            if len(self.rifs):
                self.isEcmp = True

        if self.verbose:
            sys.stderr.write('ecmpgrp %s, %s\n' % (self.nhops, self.rifs))

    def getL3PktRes(self):
        return self.isRoutable, self.isEcmp

    def getNextHopCnt(self):
        return len(self.rifs)

    def _getLagMemberFromLagLogicalId(self, lport):
        lagMembers = mlx_get_lag_member_ids(swid=0, logical_bond_id=lport)
        if not lagMembers:
            raise RuntimeError('No lag members in lag %d', lport)
        if self.verbose:
            sys.stderr.write('found lagMembers:  %s\n' % (lagMembers))

        # xlate logical port id to phy port id
        return (lagMembers[0] >> 8) & 0xff

    def _getLinuxIntfFromNhop(self, nhop, rif):
        # use nhop/rif to get neigh entry
        dstAddr = ipaddr.IPAddress(self.dst.addrstr)
        neighTable = mlx_get_neighbor(dstAddr.version, rif)
        #if self.verbose:
        #    sys.stderr.write('neighTable:\n  %s\n' % neighTable)
        dmac = ""
        # neigh format
        # 11.0.0.6 36 {'action': 'FORWARD', 'rif': 36L, 'mac': '00:02:00:00:00:06', 'vrid': 36L, 'prio': 0}
        for neigh in neighTable.items():
            if neigh[0].split()[0] == nhop:
                if self.verbose:
                    sys.stderr.write('found neigh:  %s %s\n' % (neigh[0], neigh[1]))
                dmac = neigh[1].get('mac')
                break
        if not dmac:
            raise RuntimeError('No neigh entry for %s %d', nhop, rif)

        # use rif to get bridge_id
        l3Intfs = mlx_get_interface().get('entries')
        #if self.verbose:
        #    sys.stderr.write('l3Intfs:\n  %s\n' % l3Intfs.get('entries'))
        bridge_id = -1
        # l3Intf format
        # 36 {'vlan': None, 'egress mode': None, 'swid': 0L, 'vrid': 0L, 'mtu': 1500L, 'mac': '44:38:39:00:a4:00', 'mac_addr': None, 'bridge id': 4133L, 'type': 'bridge', 'port': None}
        for l3Intf in l3Intfs.items():
            if rif == l3Intf[0]:
                if self.verbose:
                    sys.stderr.write('found l3Intf:  %s %s\n' % (l3Intf[0], l3Intf[1]))
                bridge_id = l3Intf[1].get('bridge id')
                break
        if bridge_id == -1:
            raise RuntimeError('No l3 interface for %d', rif)

        # use dmac from neigh and bridge_id from l3_interface to get the fdb entry
        fdb = mlx_get_fdb(swid = 0)
        #if self.verbose:
        #    sys.stderr.write('fdb:\n  %s\n' % fdb)
        # fdb entry format
        # 00:02:00:00:00:06 {'fid_vid': 4133L, 'swid': 0, 'logical_port': 811860224, 'lport_type': 3, 'entry_type': 'UC_AGEABLE', 'action': 'FWD', 'port_id': '1', 'vlan_id': 100}
        port = -1
        logical_port = -1
        port_type = 0
        for mac in fdb:
            if mac != dmac:
                continue
            for fdb_info in fdb[mac]:
                if fdb_info.get('fid_vid') == bridge_id and\
                    fdb_info.get('entry_type') != 'UC_AGED':
                    logical_port = fdb_info.get('logical_port')
                    port = int(fdb_info.get('port_id'))
                    port_type = fdb_info.get('lport_type')
                    if self.verbose:
                        sys.stderr.write('found fdb entry:  %s %s\n' % (mac, fdb_info))
                    break
        if port == -1:
            raise RuntimeError('No fdb entry for %d %s', bridge_id, mac)

        # use logical port from the fdb entry to get swpX
        if port_type & 1:
            # LAG
            if self.verbose:
                sys.stderr.write('lag_id: %d/%d\n' % (port, logical_port))
            port = self._getLagMemberFromLagLogicalId(logical_port)
            if self.verbose:
                sys.stderr.write('lag member: %d\n' % (port))
            linuxIntf = asictrunkmbr2linux('xe%d' % port)
        elif port_type & 2:
            # phy port
            if self.verbose:
                sys.stderr.write('phy_port: %d/%d\n' % (port, logical_port))
            linuxIntf = asic2linux('xe%d' % port)
        else:
            raise RuntimeError('logical port %d:%d has an unexpected type', port_type, port)

        return linuxIntf

    def getLinuxIntf(self, nextHopIndex):
        try:
            rif = self.rifs[nextHopIndex]
        except IndexError:
            rif = -1
        try:
            nhop = self.nhops[nextHopIndex]
        except IndexError:
            nhop = ""
        if self.verbose:
            sys.stderr.write('nhop %s, rif %d\n' % (nhop, rif))
    
        if rif == -1:
            raise RuntimeError('Error: traffic to %s will not ECMP' % self.dst)

        return self._getLinuxIntfFromNhop(nhop, rif)
        #return 'rif: %d' % rif


class HashCalc:
    _mlxHashCalcPath = '/usr/lib/cumulus/hash_calculator'
    _mlxPath = '/var/run/mlx_ecmpcalc'
    _mlxEpPath = _mlxPath + '/ep'
    _mlxPktPath = _mlxPath + '/pkt'

    def __init__(self, inIntf, frame, numNextHops, verbose=False):
        self.verbose = verbose
        self.inIntf = inIntf
        self.frame = frame

        # hardcoded hash_calc params
        self.param_et = 0x8025 # REG_ID_RECRV2
        self.param_r = 1  # ???

        # We need to input the REG_ID_RECRV2 value, but since this value
        # changes based on whether symmetric hash is enabled/disabled and
        # if the hash_seed value is changed in traffic.conf, get these
        # values from cl-cfg.
        try:
            os.makedirs(self._mlxPath)
        except OSError:
            pass
        ep3 = """    0x00 0x00 0x00 0x00
                  0x00 0x00 0x01 0xf8
                  0x00 0x00 0x00 0x00
                  0x00 0x00 0x00 0x00
                  0x00 0x00 0x0c 0x00
                  0x62 0xff 0xe0 0x3f
                  0xe0 0x2f 0xfe 0x00
                  0x00 0x00 0x00 0x00
                  0x00 0x00 0x01 0xf8
                  0x00 0x00 0xff 0xf0
                  0x1f 0xf0 0x1f 0xf8"""
        cmdStr = "/usr/cumulus/bin/cl-cfg -a switchd traffic.symmetric_hash_enable"
        try:
            sym_hash = subprocess.check_output(cmdStr.split())
        except subprocess.CalledProcessError:
            sym_hash = ""
        if self.verbose:
            sys.stderr.write('sym_hash:  %s\n' % sym_hash)
        if 'TRUE' in sym_hash :
            ep1 = """                  0x00 0x00 0x01 0x00
                  """
        else:
            ep1 = """                  0x00 0x00 0x00 0x00
                  """

        cmdStr = "/usr/cumulus/bin/cl-cfg -a switchd traffic.ecmp_hash_seed"
        try:
            hash_seed = subprocess.check_output(cmdStr.split())
        except subprocess.CalledProcessError:
            hash_seed = ""
        # extract hash seed, convert to hex and pad it to 4 byte hex-bytes
        seed = (hex(int(hash_seed.partition("=")[2]))[2:]).zfill(8)
        ep2 = """0x%s 0x%s 0x%s 0x%s
              """ % (seed[0:2], seed[2:4], seed[4:6], seed[6:8])
        if self.verbose:
            sys.stderr.write('hash_seed: %s, seed %s\n' % (hash_seed, seed))

        with open(self._mlxEpPath, 'w') as file:
            file.write(ep1)
            file.write("""0x00 0x00 0x00 0x00
                  """)
            file.write(ep2)
            file.write(ep3)

        self.nextHopIndex = -1
        self._runCalc(numNextHops)

    def _getPortMac(self):
        _macPath = '/sys/class/net/%s/address' % self.inIntf
        if not os.path.isfile(_macPath):
            raise RuntimeError('%s does not exist' % _macPath)
        try:
            with open(_macPath) as f:
                pmac = f.readline().strip()
        except IOError:
            raise RuntimeError('Contents of %s are not accessible' % _macPath)

        return pmac

    def _morphPkt(self, pkt):
        # in-packet is a hex-dump; out-pkt needs to be comma-space
        # separated hex-bytes
        pktList = [pkt[i:i+2] for i in range(0, len(pkt), 2)]
        if len(pktList):
            pktList[0] = '0x' + pktList[0]
        return ', 0x'.join(pktList)

    def _buildPkt(self):
        # we need to build a routable ethernet header; to do that -
        # 1. vlan has to be valid i.e. source port has to be a member of the
        #    vlan. Best guess would be PVID.
        # 2. dmac has to be a mystation entry. Best guess would again be the
        #    port mac.
        pmac = self._getPortMac()
        vid = self.frame.vid.value if self.frame.vid else 0
        pkt = self.frame.buildVisibilityPkt(smac=pmac, dmac=pmac, vid=vid)

        if not pkt:
            raise RuntimeError('Error: visibility pkt trace generation failed')

        # convert the hexdump into some thing that mlx hash_calc recogonizes
        pkt = self._morphPkt(pkt)
        with open(self._mlxPktPath, 'w') as file:
            file.write(pkt)

        if self.verbose:
            sys.stderr.write('Trace pkt:\n  %s\n' % pkt)

    def _runCalc(self, numNextHops):
        if not os.path.isfile(self._mlxHashCalcPath):
            raise RuntimeError('Cannot find %s' % self._mlxHashCalcPath)

        # build and write pkt to _mlxPktPath
        self._buildPkt()

        # hash_calc
        cmdStr = '%s -p %s -ep %s -s %d -o %d -et 0x%x -r %d' %\
                    (self._mlxHashCalcPath, self._mlxPktPath, 
                     self._mlxEpPath, self.frame.sportid.value, numNextHops,
                     self.param_et, self.param_r)
        if self.verbose:
            sys.stderr.write('run %s\n' % cmdStr)
        try:
            calcOut = subprocess.check_output(cmdStr.split())
        except subprocess.CalledProcessError:
            calcOut = ""
        if self.verbose:
            sys.stderr.write('calcOut:\n  %s\n' % calcOut)

        # get nexthop index
        pat = re.compile('from \d+ we picked the index port (?P<index>\d+)')
        obj = pat.search(calcOut)
        if obj:
            self.nextHopIndex = int(obj.group('index'))

        if self.verbose:
            sys.stderr.write('nexthopIndex %d\n' % self.nextHopIndex)

    def getNextHopIndex(self):
        return self.nextHopIndex


def getEcmpMbrLinuxIntf(inIntf, frame, verbose=False):
    mlx_open_connection()

    # LPM
    # XXX - need to add vrf support to ecmpcalc
    vrf = 0 
    if verbose:
        sys.stderr.write('Run ecmpcalc %s\n' % frame.dst)
    routeInfo = RouteInfo(vrf, frame.dst, verbose)
    isRoutable, isEcmp = routeInfo.getL3PktRes()
    if not isRoutable:
        raise RuntimeError('Error: traffic to %s will not be L3 switched' % frame.dst)
    if not isEcmp:
        raise RuntimeError('Error: traffic to %s will not ECMP' % frame.dst)
    numNextHops = routeInfo.getNextHopCnt()

    # nextHopIndex for a ecmp group
    hashCalc = HashCalc(inIntf, frame, numNextHops, verbose)
    nextHopIndex = hashCalc.getNextHopIndex()
    if nextHopIndex == -1:
        raise RuntimeError('hash_calc failed to locate a next hop index')

    # swpX associated with nexthop
    outIntf = routeInfo.getLinuxIntf(nextHopIndex)

    mlx_close_connection()
    return outIntf
