#
# Copyright 2016.  Cumulus Networks, Inc.
#
import os
import subprocess
import cumulus.platform
import cumulus.cpld
import cumulus.ledup
import cumulus.sunit
import cumulus.sfp
import cumulus.sensors

class FBWedge100QSFP28(cumulus.platform.Tomahawk100GPort):
    '''
    TBD
    '''
    _set_led_common = '''
;;;
;;; common LED routines
;;;
set_port_led:

        ret

set_led_off:
        ret
'''

    _set_led_1x = '''
;;;
;;; In 1x mode, only the green LED of the first bi-color LED is used to
;;; indicate link/activity. The rest are off.
;;;
set_led:
        ret
'''

    _set_led_2x = '''
;;;
;;; In 2x mode, we use the green LED of the first and second bi-color LED to
;;; indicate link/activity. The rest are off.
;;;
set_led:
        ret
'''

    _set_led_4x = '''
;;;
;;; In 4x mode, the green LEDs of all four bi-color LEDs are used to
;;; indicate link/activity.
;;;
set_led:
        ret
'''

    def __init__(self, label, serdes, rx_lane_map, tx_lane_map, leds):
        sfp = cumulus.sfp.SFPROMDTSLabel(label, 'qsfp28')
        cumulus.platform.Tomahawk100GPort.__init__(self,
                                                   label=label,
                                                   serdes=serdes,
                                                   phy_types=None,
                                                   phy_addrs=None,
                                                   phy_layer=sfp.type.upper(),
                                                   rx_lane_map=rx_lane_map,
                                                   tx_lane_map=tx_lane_map,
                                                   leds=leds,
                                                   oversub=True,
                                                   unit=0)
        self.sfp = sfp

    def set_leds(self, mode, led_port, sub, led_index_list):
        self._check_mode(mode)

        set_calls = {}

        if sub != 0:
            # For breakout modes only update LEDs for the first
            # sub-interface.
            return set_calls

        fn_name = '%s' % self.__class__.__name__
        led_port = self.led_port_num(mode, sub)
        num_lp = self.num_logical_ports(mode)
        if num_lp == 4:
            call = (fn_name + "_4x", self._set_led_4x, led_port)
        elif num_lp == 2:
            call = (fn_name + "_2x", self._set_led_2x, led_port)
        else:
            call = (fn_name + "_1x", self._set_led_1x, led_port)

        set_calls[led_index_list[0]] = call

        return set_calls

class FBWedge100LED(cumulus.ledup.TomahawkLEDController):
    led_functions = (FBWedge100QSFP28._set_led_common)
    num_leds = 128

class FBWedge100Switch(cumulus.platform.BCMSwitch):
    ledup = FBWedge100LED
    led_controllers = set([0, 1, 2])
    _qsfp28 = FBWedge100QSFP28
    has_phy_ucode_mdio = False
    has_phy_ucode_spi_rom = False

    ports = (
        # FC28-31
        _qsfp28(label="1",  serdes=(116, 117, 118, 119), rx_lane_map=0x2301, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="2",  serdes=(112, 113, 114, 115), rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="3",  serdes=(124, 125, 126, 127), rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="4",  serdes=(120, 121, 122, 123), rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),

        # FC0-3
        _qsfp28(label="5",  serdes=(4, 5, 6, 7),         rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="6",  serdes=(0, 1, 2, 3),         rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="7",  serdes=(12, 13, 14, 15),     rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="8",  serdes=(8, 9, 10, 11),       rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),

        # FC4-7
        _qsfp28(label="9",  serdes=(20, 21, 22, 23),     rx_lane_map=0x2301, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="10", serdes=(16, 17, 18, 19),     rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="11", serdes=(28, 29, 30, 31),     rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="12", serdes=(24, 25, 26, 27),     rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),

        # FC8-11
        _qsfp28(label="13", serdes=(36, 37, 38, 39),     rx_lane_map=0x2301, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="14", serdes=(32, 33, 34, 35),     rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="15", serdes=(44, 45, 46, 47),     rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="16", serdes=(40, 41, 42, 43),     rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),
        
        # FC12-15
        _qsfp28(label="17", serdes=(52, 53, 54, 55),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="18", serdes=(48, 49, 50, 51),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="19", serdes=(60, 61, 62, 63),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="20", serdes=(56, 57, 58, 59),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),

        # FC16-19
        _qsfp28(label="21", serdes=(68, 69, 70, 71),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="22", serdes=(64, 65, 66, 67),     rx_lane_map=0x3210, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="23", serdes=(76, 77, 78, 79),     rx_lane_map=0x0213, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="24", serdes=(72, 73, 74, 75),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),

        # FC20-23
        _qsfp28(label="25", serdes=(84, 85, 86, 87),     rx_lane_map=0x0123, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="26", serdes=(80, 81, 82, 83),     rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="27", serdes=(92, 93, 94, 95),     rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="28", serdes=(88, 89, 90, 91),     rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
        
        # FC24-27
        _qsfp28(label="29", serdes=(100, 101, 102, 103), rx_lane_map=0x0123, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="30", serdes=(96, 97, 98, 99),     rx_lane_map=0x1032, tx_lane_map=0x1032, leds=(0,0)),
        _qsfp28(label="31", serdes=(108, 109, 110, 111), rx_lane_map=0x1032, tx_lane_map=0x3210, leds=(0,0)),
        _qsfp28(label="32", serdes=(104, 105, 106, 107), rx_lane_map=0x3210, tx_lane_map=0x1032, leds=(0,0)),
    )

class FBWedge100(cumulus.platform.Platform):
    name = 'facebook,wedge100'
    chip = cumulus.platform.TomahawkChip()
    switch = FBWedge100Switch(chip)
    cpld = None
    status_leds = None
    pwms = None
    sensors = None

Platform = FBWedge100
